
#ifdef WIN32
#include <windows.h>    // FindFirstFile, ...
#else
#include <dirent.h>     // opendir, readdir, closedir
#endif

#include <stdarg.h>     // varargs
#include <fstream>      // ostream
#include <stdio.h>      // FILE - lexeriui
#include <iostream>

#include "utils/arocfg.h"

#include "utils.h"
#include "tclass.h"

using namespace std;

const CString btzVersion = "0.99.3 (\"when the zayac broke free\")";

/** CMD LINE args **********************************************************/

bool cfgQuiet;  // false
bool cfgVerbose;// false
bool cfgNogen;  // false
bool cfgErrors; // true
bool cfgWarns;  // true
CString cfgExt;     // "btz"
CString cfgDir;     // "bitikzyz" : prefix
CString cfgIncl;    // "../bitikzyz/vmachine.h"
CString cfgParent;  // ""
CString cfgParIncl; // ""
CString cfgNamesp;  // ""
char *cfgLang;      // "languageLT"
char *cfgVoid;      // "void"
char *cfgLong;      // "long"
char *cfgFloat;     // "float"

CString cfgOut;   // prefix+".dat"
CString cfgH;     // prefix+".h"
CString cfgCPP;   // prefix+".cpp"
CString cfgRoot;  // prefix+"Root"

char *argHelp[]    = {"--help",     "-h" };
char *argVersion[] = {"--version",  NULL };
char *argQuiet[]   = {"--quiet",    "-q" };
char *argVerbose[] = {"--verbose",  "-v" };
char *argNogen[]   = {"--no-gen",   "-n" };
char *argErrors[]  = {"--no-err",   "-e" };
char *argWarns[]   = {"--no-warn",  "-w" };
char *argExt[]     = {"--ext",      "-x" };
char *argDir[]     = {"--dir",      "-d" };
char *argIncl[]    = {"--include",  "-i" };
char *argParent[]  = {"--parent",   "-p" };
char *argParIncl[] = {"--par-incl", NULL };
char *argNamesp[]  = {"--namespace",NULL };

bool config(long argc, char **argv)
{
        char *s;
        TArguments a(argc, argv);
        if (a.is(argHelp)) {
            cout<<strHelpUsage<<'\n';
            cout<<strHelpKeys <<'\n';
            cout<<strHelpLine1<<'\n';
            cout<<strHelpLine2<<'\n';
            cout<<strHelpLine3<<'\n';
            cout<<strHelpLine4<<'\n';
            cout<<strHelpLine5<<'\n';
            cout<<strHelpLine6<<'\n';
            cout<<strHelpLine7<<'\n';
            cout<<strHelpLine8<<'\n';
            cout<<strHelpLine9<<'\n';
            cout<<strHelpLineA<<'\n';
            cout<<strHelpLineB<<'\n';
            cout<<strHelpLineC<<'\n';
            cout<<strHelpLineD<<'\n';
            cout<<strHelpDef1 <<'\n';
            cout<<strHelpDef2 <<'\n';
            return false;
        }
        if (a.is(argVersion)) {
            cout<<strHelpVer1<<btzVersion<<'\n';
            cout<<strHelpVer2<<'\n';
            return false;
        }
        cfgQuiet  = a.is(argQuiet)  ? true : (bool)cfgGetL("options", "quiet", 0);
        cfgVerbose= a.is(argVerbose)? true : (bool)cfgGetL("options", "verbose", 0);
        cfgNogen  = a.is(argNogen)  ? true : (bool)cfgGetL("options", "nogen", 0);
        cfgErrors = a.is(argErrors) ? false: (bool)cfgGetL("options", "errors", 1);
        cfgWarns  = a.is(argWarns)  ? false: (bool)cfgGetL("options", "warnings", 1);
        s = a.findS(argExt);    cfgExt  = s ? s : cfgGetS("options", "extension" ,".btz");
        s = a.findS(argDir);    cfgDir  = s ? s : cfgGetS("options", "directory" ,"bitikzyz");
        while (cfgDir[cfgDir.size()-1] == '/') cfgDir.erase(1);
        s = a.findS(argParent);  cfgParent  = s ? s : cfgGetS("options", "parent" ,"");
        s = a.findS(argParIncl); cfgParIncl = s ? s : cfgGetS("options", "par-include" ,"");
        s = a.findS(argIncl);    cfgIncl    = s ? s : cfgGetS("options", "include" ,"../bitikzyz/vmachine.h");
        s = a.findS(argNamesp);  cfgNamesp  = s ? s : cfgGetS("options", "namespace" ,"");
        // tik is konfigo
        cfgVoid  = cfgGetS("options", "type-void" ,"void");
        cfgLong  = cfgGetS("options", "type-long" ,"long");
        cfgFloat = cfgGetS("options", "type-float" ,"float");
        // patys uzpildom
        cfgOut = cfgDir + ".dat";
        cfgH   = cfgDir + ".h";
        cfgCPP = cfgDir + ".cpp";
        cfgRoot= cfgDir + "Root";
        return true;
}

/** TDirectory *************************************************************/

class TDirectory {
        char *dir;
        long n;
        // grazus kodas...
        #ifdef WIN32
        struct TDumbStruct { char *d_name; };
        TDumbStruct *files;
        #else
        dirent *files;
        #endif
public:
        TDirectory(const char *d = "./");
        ~TDirectory();
        long num() { return n; };
        char *name(long i) { return files[i].d_name; };
};

TDirectory::TDirectory(const char *d)
{
        n = 0;
        files = NULL;
        dir = new char[strlen(d)+1];
        strcpy(dir, d);
#ifdef WIN32
        char buf[256];
        strcpy(buf, d);
        strcpy(buf + strlen(d), "*.*");
        // ziurim, kiek failu is viso
        WIN32_FIND_DATA ff;
        HANDLE h = FindFirstFile(buf, &ff);
        int fl = (h != INVALID_HANDLE_VALUE);
        while (fl) {
                n++;
                fl = FindNextFile(h, &ff);
        }
        FindClose(h);
        // isskiriam masyva vardams
        if (!n) return;
        files = new TDumbStruct[n];
        // surasom vardus
        // **** HACK -  kazkodel suporolina... kodel? windowsai?
        strcpy(buf, d);
        strcpy(buf + strlen(d), "*.*");
        // ***** baigesi hackas
        h = FindFirstFile(buf, &ff);
        fl = (h != INVALID_HANDLE_VALUE);
        long i = 0;
        while (fl) {
                files[i].d_name = new char[strlen(ff.cFileName)+1];
                strcpy(files[i].d_name, ff.cFileName);
                fl = FindNextFile(h, &ff);
                i++;
        }
        FindClose(h);
#else
        dir = new char[strlen(d)+1];
        strcpy(dir, d);
        // POSIX
        DIR *p = opendir(d);
        if (!p) return;
        while (readdir(p)) n++;
        rewinddir(p);
        files = new dirent[n];
        for (long i = 0; i < n; i++) {
                dirent *pp = readdir(p);
                if (!pp) {
                        n = i;
                        break;
                }
                memcpy(&files[i], pp, sizeof(dirent));
        }
        closedir(p);
#endif
}

TDirectory::~TDirectory()
{
#ifndef WIN32
        for (long i = n; i < n; i++)
                delete[] files[i].d_name;
#endif
        if (files) delete[] files;
        delete[] dir;
        files = NULL;
        dir = NULL;
        n = 0;
}

/** flex/bison *************************************************************/

extern long lineNumber; // flex - eil. nr.
extern FILE *yyin; // flex - input stream
int yylex();   // flex - main f-ja
int yyparse(); // bison - main f-ja

void lexerShutDown();

// sita kviecia bison
void yyerror(char *msg) { warnMsg("bison: %s", (char*)msg); }
// sita kviecia flex, kai baigiasi failas
extern "C" int yywrap(void) { return 1; }
// bison
TClass *currClass;  // dabar kompil. klase
long    staticSize; // statinio duom. segmento dydis
long    nativeSlots; // kiek native f-ju is viso yra
extern long parseMode; // pass# (t.y. 1, 2 arba 3)

/***************************************************************************/
/* raportuoja klaidas */

long numErrors = 0;

void warnMsg(const char *format, ...)
{
        if (!cfgWarns)
                return;
        va_list args;
        printf("WRN %s %li: ", currClass->Name().c_str(), lineNumber);
        va_start(args, format);
        vfprintf(stdout, format, args);
        va_end(args);
        printf("\n");
}
void errorMsg(const char *format, ...)
{
        numErrors++;
        if (!cfgErrors)
                return;
        va_list args;
        printf("ERR %s %li: ", currClass->Name().c_str(), lineNumber);
        va_start(args, format);
        vfprintf(stdout, format, args);
        va_end(args);
        printf("\n");
}

/** KOMPILIAVIMAS **********************************************************/

TClass *root;

void compileDir(CString& n, TClass *p);

/** PASS#1: diru strukturos skanavimas - klasiu sukurimas ******************/

void compileFile1(CString& fn, CString& cn, TClass *p)
{
        // atidarom faila
        yyin = fopen(fn.c_str(), "rt");  // lexerio inputas
        if (!yyin) return;
        // sukuriam klase
        TClass *c = new TClass(cn, p);
        p->add(c);
        // skanuojam vaikines
        fn.erase(cfgExt.size());
        compileDir(fn, c);
        return;
}

/** PASS#2: tipu, kintamuju ir funkciju skanavimas *************************/

void compileFile2(CString& fn, CString& cn, TClass *p)
{
        // atidarom faila
        yyin = fopen(fn.c_str(), "rt");  // lexerio inputas
        if (!yyin) return;
        lineNumber = 1;
        // randam klase
        TClass *c = p->findName(cn);
        // sureguliuojam offsetus
        c->Vars()->Offset ( - p->Vars()->Size() + p->Vars()->Offset() );
        c->Funcs()->Base  ( p->Funcs()->Size() + p->Funcs()->Base() );
        // gerai
        currClass = c;
        yyparse(); // kompiliuojam (parseMode=2)
        fclose(yyin);
        // kompiliuojam vaikines
        fn.erase(cfgExt.size());
        compileDir(fn, c);
        return;
}

/** PASS#3: kompiliavimas **************************************************/

void compileFile3(CString& fn, CString& cn, TClass *p)
{
        // atidarom faila
        yyin = fopen(fn.c_str(), "rt");  // lexerio inputas
        if (!yyin) return;
        lineNumber = 1;
        // randam klase
        TClass *c = p->findName(cn);
        currClass = c;
        yyparse(); // kompiliuojam (parseMode=3)
        fclose(yyin);
        c->Funcs()->code(); // darom tarpini koda, ji optimizuojam
        if (!cfgQuiet && cfgVerbose) {
                cout<<"   "<<cn<<":\n";
                c->Funcs()->debug(2); // debuginam
        }
        // kompiliuojam vaikines
        fn.erase(cfgExt.size());
        compileDir(fn, c);
        return;
}

/** viso diro kompil *******************************************************/

void compileDir(CString& n, TClass *p)
{
        // tvarkom varda - gal gale reikia '/'
        if ( n[n.size()-1] != '/' ) n += '/';
        // diras
        TDirectory d(n.c_str());
        CString fname;
        CString cname;
        // begam per visus failus
        for (long i = 0; i < d.num(); i++) {
                char *s = d.name(i);
                long slen = strlen(s);
                long elen = cfgExt.size();
                if (slen < elen) continue;
                if (strcmp(s + slen - elen, cfgExt.c_str())) continue;
                fname = n + s;
                cname = s;
                cname.erase(elen);
                switch (parseMode) {
                case 1: compileFile1(fname, cname, p); break;
                case 2: compileFile2(fname, cname, p); break;
                case 3: compileFile3(fname, cname, p); break;
                }
        }
}

/***************************************************************************/

int main(int argc, char **argv)
{
        cfgFile("default.cfg");
        cfgLang = cfgGetS("options", "language", "languageLT");
        getStrings(); // nes jam reikia cfgLang
        // config
        if (!config(argc, argv))
                return 0;
        // create root
        if (!cfgQuiet) cout<<strCompRoot<<cfgRoot<<'\n';
        root = new TClass(cfgRoot);
        // static ir native: tusti
        staticSize = 0;
        nativeSlots = 0;
        // pass #1
        CString dir = cfgDir;
        if (!cfgQuiet) cout<<strComp1<<'\n';
        parseMode = 1;  compileDir(dir, root);
        //m_dumpMemoryReport(); // MMGR
        // pass #2
        if (!cfgQuiet) cout<<strComp2<<'\n';
        parseMode = 2;  compileDir(dir, root);
        // pass #3
        if (!cfgQuiet) cout<<strComp3<<'\n';
        parseMode = 3;  compileDir(dir, root);
        // asm; write code
        if (!cfgQuiet) cout<<strCompOut<<cfgOut<<'\n';
        root->save(cfgOut);
        if (!cfgNogen) {
            // gen C++ interface
            if (!cfgQuiet) cout<<strCompCPP<<cfgCPP<<", "<<cfgH<<'\n';
            // open H
            ofstream fh (cfgH.c_str());
            // header
            fh<<"// BiTikZyz "<<btzVersion<<'\n';
            fh<<strCompHStart<<'\n';
            fh<<"#ifndef __btz_"<<cfgDir<<"_h__\n";
            fh<<"#define __btz_"<<cfgDir<<"_h__\n";
            fh<<"#include \""<<cfgIncl<<"\"\n";
            if (cfgParIncl[0]) fh<<"#include \""<<cfgParIncl<<"\"\n";
            fh<<"#define MANO_PI 3.1415926535897932384626433\n";
            if (cfgNamesp[0]) fh<<"namespace "<<cfgNamesp<<" {\n";
            fh<<"extern TVMachine "<<cfgDir<<";\n";
            // forwards
            fh<<"// **** forward decls ****\n";
            root->printFwd(fh);
            fh<<"// **** definitions ****\n";
            root->printDef(fh);
            fh<<"// **** inlines ****\n";
            root->printInl(fh);
            // footer
            if (cfgNamesp[0]) fh<<"\n}\n\n";
            fh<<"#endif\n\n";
            // gen C++ source
            ofstream fcpp (cfgCPP.c_str());
            fcpp<<"// BiTikZyz "<<btzVersion<<'\n';
            fcpp<<strCompCPPStart<<'\n';
            fcpp<<"#include <math.h>\n";
            fcpp<<"#include <stdlib.h>\n";
            fcpp<<"#include <stdio.h>\n";
            fcpp<<"#include <string.h>\n";
            fcpp<<"#include \""<<cfgH<<"\"\n";
            if (cfgNamesp[0]) fcpp<<"using namespace "<<cfgNamesp<<";\n"
                                  <<"namespace "<<cfgNamesp<<" {\n";
            fcpp<<"long "<<cfgDir<<"_static["
                <<(staticSize ? staticSize : staticSize + 1)<<"];\n";
            root->printCpp(fcpp);
            root->printNat(fcpp);
            fcpp<<"TVMNative "<<cfgDir<<"_native["
                <<(nativeSlots ? nativeSlots : nativeSlots + 1)<<"] = {\n    ";
            long i, j;
            for (i = 0, j = 0; i < nativeSlots; i++) {
                    fcpp<<cfgDir<<"_native_"<<i<<", ";
                    if (++j > 3) {
                            j = 0;
                            fcpp<<"\n    ";
                    }
            }
            fcpp<<"};\n";
            fcpp<<"TVMachine "<<cfgDir<<"(\""<<cfgOut<<"\", (unsigned char*)"<<cfgDir
                <<"_static, "<<cfgDir<<"_native);\n\n";
            if (cfgNamesp[0]) fcpp<<"\n}\n\n";
        }
        // end
        if (!cfgQuiet || numErrors) cout<<strCompEnd<<numErrors<<'\n';
        delete root;
        cfgClose();
        lexerShutDown();
        return numErrors ? 1 : 0;
}

