// visu TNode optimize(), remove() ir type() funkcijos
// jos iterpia ir reikalingus konversijos nodus bei patikrina semantika.
// jei kas nors blogai, tai vietoj shakos padaro nuli arba tuscia shaka.

#include <iostream>
#include <stdio.h>

#ifdef MMGR
#include "../utils/mmgr.h"
#endif

#include "syntree.h"
#include "utils.h"

#define ISREF(ttt) ((ttt)==TypeStr || (ttt)->Cla())

TNode *TNode::matchType( TType *t, TNode *n, const char *descr )
{
	TType *ntype = n->type();
	if( t == ntype )
		return n;
	// check std. coerces
	if( t == TypeLong && ntype == TypeFloat ) {
		TNode *nn = new T2Long( n );
		nn = nn->optimize();
		return nn;
	}
	if( t == TypeFloat && ntype == TypeLong ) {
		TNode *nn = new T2Float( n );
		nn = nn->optimize();
		return nn;
	}
	if( t == TypeLong && ISREF(ntype) ) {
		warnMsg( strWarnPtr.c_str(), "multi" );
		return n;
	}
	// check class refs
	if( t->Cla() && ntype->Cla() ) {
		while( ntype->Cla()->Parent() ) {
			ntype = ntype->Cla()->Parent()->Type();
			if( t == ntype )
				return n;
		}
	}
	// check multi-expr
	if( t->Memb() && isMultiExpr( n ) ) {
		TMultiExprNode *m = isMultiExpr( n );
		TSymTable *s = t->Memb();
		if( m->Count() != s->Size() ) {
			m->del(); return new TEmptyNode;
		}
		// suzhiurim arg tipus
		TTableIterator it = s->items();
		TArgs *a = m->Exprs();
		while( a ) {
			TSymbol *s = (TSymbol*)it.next();
			TType *t1 = s->Type();
			if( !a->expr ) {
				errorMsg( strErrBadArg.c_str(), descr );
				m->del(); return new TEmptyNode;
			}
			TNode *coerced = matchType( t1, a->expr, descr );
			if( isEmpty( coerced ) ) {
				errorMsg(strErrBadTypes.c_str(), descr );
				m->del(); return new TEmptyNode;
			}
			a->expr = coerced;
			a = a->next;
		}
		return m;
	}
	// bad
	errorMsg(strErrBadTypes.c_str(), descr );
	n->del(); return new TEmptyNode;
}


TNode *TListNode::optimize()
{
	TNode *n = remove();
	return n;
}

TNode *TListNode::remove()
{
	node = node->remove();
	list = list->remove();
	TEmptyNode *en = isEmpty(node);
	TEmptyNode *el = isEmpty(list);
	if (en) {
		TNode *p = list;
		delete en;
		delete this;
		return p;
	} else if (el) {
		TNode *p = node;
		delete el;
		delete this;
		return p;
	}
	return this;
}

TNode *TCondStatNode::optimize()
{
	TType *t = cond->type();
	if (t == TypeLong || ISREF(t))
		return this;
	if (t == TypeFloat) {
		cond = new T2Long(cond);
		return this;
	}
	// blogai
	errorMsg(strErrNeedExpr.c_str(), "condstat");
	del();
	return new TEmptyNode();
}

TNode *TTernaryNode::optimize()
{
	// tipu konversijos
	TType *t = type();
	if (t == TypeError) {
		errorMsg(strErrBadTypes.c_str(), "ternary");
		del();
		return new TNumberNode(0);
	}
	stat = matchType( t, stat, "ternary" );
	stat2 = matchType( t, stat2, "ternary" );
	// grazinam sita
	return this;
}

TType *TTernaryNode::type() const
{
	// tipu konversijos
	TType *t1 = stat->type();
	TType *t2 = stat2->type();
	if (t1 == t2) return t1;
	if (t1->Cla() && t2->Cla()) {
		TType *t = (t1->Cla()->Level() > t2->Cla()->Level()) ? t2 : t1;
		return t;
	}
	if (ISREF(t1) && t2 == TypeLong) return t1;
	if (ISREF(t2) && t1 == TypeLong) return t2;
	if (t1 == TypeLong && t2 == TypeFloat) return TypeFloat;
	if (t2 == TypeLong && t1 == TypeFloat) return TypeFloat;
	return TypeError;
}

TNode *TForNode::optimize()
{
	TType *t = cond->type();
	if (t == TypeLong || ISREF(t))
		return this;
	if (t == TypeFloat) {
		cond = new T2Long(cond);
		cond = cond->optimize();
		return this;
	}
	// blogai
	errorMsg(strErrNeedExpr.c_str(), "for");
	del();
	return new TEmptyNode();
}

/******************
 * BINARINES ISHR *
 ******************/

TNode *TBinaryNode::optimize()
{
	TType *t = type(); // patikrinam tipa
	if (t == TypeError) { // klaida
		del();
		errorMsg(strErrBadTypes.c_str(), "binary");
		return new TNumberNode(0);
	}
	// tipu konversijos
	TType *t1 = op1->type();
	TType *t2 = op2->type();
	if (t == TypeFloat) { // rezultas turi buti float
		if (t1 == TypeLong) {
				op1 = new T2Float(op1);
				op1 = op1->optimize();
		}
		if (t2 == TypeLong) {
				op2 = new T2Float(op2);
				op2 = op2->optimize();
		}
	} else if (!p->lng || (t1 != t2)) { // rezultas turi buti long
		if (t1 == TypeFloat) {
				op1 = new T2Long(op1);
				op1 = op1->optimize();
		}
		if (t2 == TypeFloat) {
				op2 = new T2Long(op2);
				op2 = op2->optimize();
		}
	}
	// gal const expr?
	TNumericNode *n1 = isNumeric(op1);
	TNumericNode *n2 = isNumeric(op2);
	if (n1 && n2) {
		TNumericNode *nn;
		float val = p->opti(n1->Val(), n2->Val());
		if (t == TypeLong)
				nn = new TNumberNode((long)val);
		else
				nn = new TFloatNode(val);
		del();
		return nn;
	}
	// grazinam sita
	return this;
}

TNode *TBinaryNode::remove()
{
		TNode *n = new TListNode(op2, op1);
		delete this;
		return n->remove();
}

TType *TBinaryNode::type() const
{
	TType *t1 = op1->type();
	TType *t2 = op2->type();
	// gal galima ref
	if (p->ref) {
		TNumericNode *n = NULL; // kad warningo nebutu
		if ((ISREF(t1) && (n = isNumeric(op2))) ||
			(ISREF(t2) && (n = isNumeric(op1)))) {
				if ((long)n->Val())
						warnMsg(strWarnNumPtr.c_str(), "binary");
				return TypeLong; // ref ir skaicius
		}
		if (ISREF(t1) || ISREF(t2)) // bent vienas ref
				return TypeLong;
	}
	// jei bent 1 ne long | float - klaida
	if (t1!=TypeLong && t1!=TypeFloat || t2!=TypeLong && t2!=TypeFloat) {
		warnMsg(strErrNeedScal.c_str(), "binary");
		return TypeError;
	}
	// jei galima float ir bent 1 float - tai float
	if (p->flt && !p->lng) {
		if (t1 == TypeFloat || t2 == TypeFloat)
				return TypeFloat;
	}
	return TypeLong;
}

/****************
 * UNARINES OP. *
 ****************/

TNode *TUnaryNode::optimize()
{
	TType *t = type(); // patikrinam tipa
	if (t == TypeError) { // klaida
		del();
		errorMsg(strErrBadTypes.c_str(), "unary");
		return new TNumberNode(0);
	}
	TType *t1 = op->type();
	// tipu konversijos
	if (t == TypeFloat) { // rezultas turi buti float
		if (t1 == TypeLong) {
				op = new T2Float(op);
				op = op->optimize();
		}
	} else if (!p->lng) { // rezultas turi buti long
		if (t1 == TypeFloat) {
				op = new T2Long(op);
				op = op->optimize();
		}
	}
	// gal const expr?
	TNumericNode *n = isNumeric(op);
	if (n) {
		TNumericNode *nn;
		float val = p->opti(n->Val());
		if (t == TypeLong)
			nn = new TNumberNode((long)val);
		else
			nn = new TFloatNode(val);
		del();
		return nn;
	}
	// grazinam sita
	return this;
}

TNode *TUnaryNode::remove()
{
	TNode *n = op;
	delete this;
	return n->remove();
}

TType *TUnaryNode::type() const
{
	TType *t = op->type();
	// gal galima ref
	if (p->ref && ISREF(t)) return TypeLong;
	// jei ne long | float - klaida
	if (t!=TypeLong && t!=TypeFloat) {
		warnMsg(strErrNeedScal.c_str(), "unary");
		return TypeError;
	}
	// jei galima float ir op float - tai float
	if (p->flt && !p->lng) return t;
	return TypeLong;
}

/******************
 * F-JU KVIETIMAS *
 ******************/

TNode *TFuncNode::optimize()
{
	// suzhiurim
	if (count != func->Args()) {
		errorMsg(strErrNeedArgs.c_str(), "call", func->Args());
		del(); return new TEmptyNode;
	}
	// suzhiurim arg tipus (praleidzhiam returnval)
	TTableIterator it = func->Tab()->items();
	it.next();
	TArgs *a = args;
	while (a) {
		TSymbol *s = (TSymbol*)it.next();
		TType *t1 = s->Type();
		if (!a->expr) {
			errorMsg(strErrBadArg.c_str(), "call");
			del(); return new TEmptyNode;
		}
		TNode *coerced = matchType( t1, a->expr, "call" );
		if( isEmpty( coerced ) ) {
			errorMsg(strErrBadTypes.c_str(), "call");
			del(); return new TEmptyNode;
		}
		a->expr = coerced;
		a = a->next;
	}
	return this;
}

/******************
 * PRISKYRIMO OP. *
 ******************/

TNode *TAsArithNode::optimize()
{
	TType *t = type(); // patikrinam tipa
	if (t == TypeError) { // klaida
		del();
		errorMsg(strErrBadTypes.c_str(), "assign");
		return new TNumberNode(0);
	}
	TType *t1 = lval->type();
	TType *t2 = expr->type();
	if (t1 == t2)
		return this;
	// tipu konversijos
	if (t1 == TypeFloat) { // rezultas turi buti float
		expr = new T2Float(expr);
		expr = expr->optimize();
	} else { // rezultas turi buti long
		expr = new T2Long(expr);
		expr = expr->optimize();
	}
	// grazinam sita
	return this;
}

TType *TAsArithNode::type() const
{
	TType *t1 = lval->type();
	TType *t2 = expr->type();
	// jei bent 1 ne long | float - klaida
	if (t1!=TypeLong && t1!=TypeFloat || t2!=TypeLong && t2!=TypeFloat) {
		warnMsg(strErrNeedScal.c_str(), "assign");
		return TypeError;
	}
	// jei galima float ir bent 1 float - tai float
	if (t1 == TypeFloat)
		return (p->flt) ? TypeFloat : TypeError;
	return TypeLong;
}

// paprastas priskyrimas
TNode *TAssignNode::optimize()
{
	TType *t1 = lval->type();
	// check spec case: ptr = int
	if( ISREF(t1) ) {
		TType *t2 = expr->type();
		if( t1 != t2 ) {
			TNumericNode *n = isNumeric(expr);
			if (!n || (n && (long)n->Val()))
				warnMsg(strWarnNumPtr.c_str(), "assign");
			if (t2 == TypeFloat) {
				expr = new T2Long(expr);
				expr = expr->optimize();
			}
		}
		return this;
	}
	// ordinary case
	expr = matchType( t1, expr, "assign" );
	return this;
}

/***************
 * KONVERSIJOS *
 ***************/

TNode *TCoerceNode::remove()
{
	TNode *n = op;
	delete this;
	return n->remove();
}

TNode *T2Long::optimize()
{
	TNumericNode *n1 = isNumeric(op);
	if (n1) {
		TNumericNode *nn;
		float val = n1->Val();
		nn = new TNumberNode((long)val);
		del();
		return nn;
	}
	return this;
}
TNode *T2Float::optimize()
{
	TNumericNode *n1 = isNumeric(op);
	if (n1) {
		TNumericNode *nn;
		float val = n1->Val();
		nn = new TFloatNode(val);
		del();
		return nn;
	}
	return this;
}

/***************************************************************************/
/* IDENTAI */

TNode *TOffsetNode::optimize()
{
		TType *t = type();
		t = expr->type();
		if (t != TypeLong && t != TypeFloat) {
				errorMsg(strErrNeedScal.c_str(), "offset");
				del(); return new TNumberNode(0);
		}
		// kad long tipas butu
		if (t != TypeLong) {
				expr = new T2Long(expr);
				expr = expr->optimize();
		}
		return this;
}
TNode *TOffsetNode::remove()
{
		TNode *n = new TListNode(expr, lval); // lval; expr;
		delete this;
		return n->remove();
}

TNode *TMemberNode::remove()
{
		TNode *n = lval;
		delete this;
		return n->remove();
}

TNode *TIdentNode::remove()
{
		TNode *n = lval;
		delete this;
		return n->remove();
}

TNode *TIsNode::remove()
{
		TNode *n = lval;
		delete this;
		return n->remove();
}

