
#include <stdio.h>
#include <iostream>

#include "syntree.h"

TFuncNode::TFuncNode(TNode *l, TFunction *f, TArgs *a, long s)
{
        lval = l;
        func = f;
        args = a;
        count = 0;
        super = s;
        while (a) {
                a = a->next;
                count++;
        }
}

TFuncNode::~TFuncNode()
{
        TArgs *a1, *a2;
        a1 = args;
        while (a1) {
                a2 = a1->next;
                delete a1;
                a1 = a2;
        }
}

/** Multi-expr ********************************************************/

TMultiExprNode::TMultiExprNode(TArgs *e)
{
	exprs = e;
	count = 0;
	while (e) {
		e = e->next;
		count++;
	}
}

TMultiExprNode::~TMultiExprNode()
{
	TArgs *a1, *a2;
	a1 = exprs;
	while (a1) {
		a2 = a1->next;
		delete a1;
		a1 = a2;
	}
}

long TMultiExprNode::TotalSize() const
{
	TArgs *e = exprs;
	long size = 0;
	while (e) {
		size += e->expr->type()->Size();
		e = e->next;
	}
	return size;
}

/** TBinaryNode ************************************************************/

float optiAdd (float v1, float v2)  { return v1 + v2; }
float optiSub (float v1, float v2)  { return v1 - v2; }
float optiMul (float v1, float v2)  { return v1 * v2; }
float optiDiv (float v1, float v2)  { return v1 / v2; }
float optiMod (float v1, float v2)  { return (long)v1 %  (long)v2; }
float optiShl (float v1, float v2)  { return (long)v1 << (long)v2; }
float optiShr (float v1, float v2)  { return (long)v1 >> (long)v2; }
float optiOr  (float v1, float v2)  { return (long)v1 |  (long)v2; }
float optiXor (float v1, float v2)  { return (long)v1 ^  (long)v2; }
float optiAnd (float v1, float v2)  { return (long)v1 &  (long)v2; }
float optiBOr (float v1, float v2)  { return v1 || v2; }
float optiBAnd(float v1, float v2)  { return v1 && v2; }
float optiBE  (float v1, float v2)  { return v1 == v2; }
float optiBNE (float v1, float v2)  { return v1 != v2; }
float optiBL  (float v1, float v2)  { return v1 <  v2; }
float optiBG  (float v1, float v2)  { return v1 >  v2; }
float optiBLE (float v1, float v2)  { return v1 <= v2; }
float optiBGE (float v1, float v2)  { return v1 >= v2; }

TBinaryNode::TParams TBinaryNode::table[] = {
    { "+"  ,IOP_ADDi  ,IOP_ADDf  ,optiAdd   ,1 ,0 ,0 },
    { "-"  ,IOP_SUBi  ,IOP_SUBf  ,optiSub   ,1 ,0 ,0 },
    { "*"  ,IOP_MULi  ,IOP_MULf  ,optiMul   ,1 ,0 ,0 },
    { "/"  ,IOP_DIVi  ,IOP_DIVf  ,optiDiv   ,1 ,0 ,0 },
    { "%"  ,IOP_MOD   ,IOP_NOP   ,optiMod   ,0 ,0 ,0 },
    { "<<" ,IOP_SHL   ,IOP_NOP   ,optiShl   ,0 ,0 ,0 },
    { ">>" ,IOP_SHR   ,IOP_NOP   ,optiShr   ,0 ,0 ,0 },
    { "|"  ,IOP_OR    ,IOP_NOP   ,optiOr    ,0 ,0 ,0 },
    { "^"  ,IOP_XOR   ,IOP_NOP   ,optiXor   ,0 ,0 ,0 },
    { "&"  ,IOP_AND   ,IOP_NOP   ,optiAnd   ,0 ,0 ,0 },
    { "||" ,IOP_BORi  ,IOP_BORf  ,optiBOr   ,1 ,1 ,1 },
    { "&&" ,IOP_BANDi ,IOP_BANDf ,optiBAnd  ,1 ,1 ,1 },
    { "==" ,IOP_BEi   ,IOP_BEf   ,optiBE    ,1 ,1 ,1 },
    { "!=" ,IOP_BNEi  ,IOP_BNEf  ,optiBNE   ,1 ,1 ,1 },
    { "<"  ,IOP_BLi   ,IOP_BLf   ,optiBL    ,1 ,0 ,1 },
    { ">"  ,IOP_BGi   ,IOP_BGf   ,optiBG    ,1 ,0 ,1 },
    { "<=" ,IOP_BLEi  ,IOP_BLEf  ,optiBLE   ,1 ,0 ,1 },
    { ">=" ,IOP_BGEi  ,IOP_BGEf  ,optiBGE   ,1 ,0 ,1 },
};

/** TUnaryNode *************************************************************/

float optiMinus(float v) { return -v; }
float optiBNot (float v) { return !v; }
float optiInv  (float v) { return ~(long)v; }
float optiU    (float v) { return v; }

TUnaryNode::TParams TUnaryNode::table[] = {
    { "pre++"  ,IOP_PREINCi  ,IOP_PREINCf  ,optiU     ,1 ,0 ,0 },
    { "pre--"  ,IOP_PREDECi  ,IOP_PREDECf  ,optiU     ,1 ,0 ,0 },
    { "post++" ,IOP_POSINCi  ,IOP_POSINCf  ,optiU     ,1 ,0 ,0 },
    { "post--" ,IOP_POSDECi  ,IOP_POSDECf  ,optiU     ,1 ,0 ,0 },
    { "-"      ,IOP_NEGi     ,IOP_NEGf     ,optiMinus ,1 ,0 ,0 },
    { "!"      ,IOP_BNOTi    ,IOP_BNOTf    ,optiBNot  ,1 ,1 ,1 },
    { "~"      ,IOP_INV      ,IOP_NOP      ,optiInv   ,0 ,0 ,0 },
};

/** TAsArithNode ***********************************************************/

TAsArithNode::TParams TAsArithNode::table[] = {
    { "*="  ,IOP_SAVEMULi  ,IOP_SAVEMULf  ,1 },
    { "/="  ,IOP_SAVEDIVi  ,IOP_SAVEDIVf  ,1 },
    { "%="  ,IOP_SAVEMOD   ,IOP_NOP       ,0 },
    { "+="  ,IOP_SAVEADDi  ,IOP_SAVEADDf  ,1 },
    { "-="  ,IOP_SAVESUBi  ,IOP_SAVESUBf  ,1 },
    { "<<=" ,IOP_SAVESHL   ,IOP_NOP       ,0 },
    { ">>=" ,IOP_SAVESHR   ,IOP_NOP       ,0 },
    { "|="  ,IOP_SAVEOR    ,IOP_NOP       ,0 },
    { "^="  ,IOP_SAVEXOR   ,IOP_NOP       ,0 },
    { "&="  ,IOP_SAVEAND   ,IOP_NOP       ,0 },
};

