/* klasiu info: hierarchija, duomenys, metodai, tipai */
// reikia <iostream>

// interfeisas:
// TClass, TType, TTypeTable, TSymbol, TSymTable, TFunction, TFunctable

#ifndef __TCLASS_H__
#define __TCLASS_H__

#include "utils/tipai.h"
#include "utils/CString.h"

#include "collection.h"

struct TString;
class TType;
class TSymbol;
class TFunction;

class TStringTable;
class TTypeTable;
class TSymTable;
class TFuncTable;

class TNode;
struct TIntInsn;

void errorMsg(const char *format, ...);
void  warnMsg(const char *format, ...);

/***************************************************************************/
/* TClass */

// laiko visa klases informacija
class TClass {
public:
        TClass(const CString& n);        // root klasei
        TClass(const CString& n, TClass *p);
        ~TClass();
        // metodai
        void       add(TClass *c);    // prideda vaikine klase
        void       printFwd(std::ostream& f); // ishveda c++ fwd
        void       printDef(std::ostream& f); // ishveda c++ def
        void       printInl(std::ostream& f); // ishveda c++ inline
        void       printCpp(std::ostream& f); // ishveda c++ cpp
        void       printNat(std::ostream& f); // ishveda c++ native
        long      *vtbl(long& sz);       // padaro vtbl
        void       save(const CString& n); // irasho VM viska i faila n
        TClass    *findId  (long i)    const;        // randa klase su num
        TClass    *findName(const CString& n) const; // randa klase su vardu
        TType     *findType(const CString& n) const; // randa
        TSymbol   *findSym (const CString& n) const;
        TFunction *findFunc(const CString& n) const;
        void addType(TType     *p); // prideda
        void addSym (TSymbol   *p);
        void addFunc(TFunction *p);
        long Size() const;          // grazina klases objekto dydi
        // read
        const CString& Name()const { return name; };
        long        Id()     const { return id; };
        TTypeTable *Types()  const { return types; };
        TSymTable  *Vars()   const { return vars; };
        TFuncTable *Funcs()  const { return funcs; };
        TType      *Type()   const { return type; };
        TClass     *First()  const { return first; };
        TClass     *Last()   const { return last; };
        TClass     *Parent() const { return parent; };
        TClass     *Next()   const { return next; };
        long        Level()  const { return level; };
        static TStringTable& Strings() { return strings; };
        // write
        void Next(TClass *c)    { next = c; };
private:
        // klases info
        CString     name;       // klases vardas
        long        id;         // numeras (id)
        TTypeTable *types;      // klases tipai
        TSymTable  *vars;       // duomenys
        TFuncTable *funcs;      // metodai
        TType      *type;       // "savo" tipas
        // klasiu hierarchija
        TClass *first;  // vaikiniu klasiu sarashas
        TClass *last;
        TClass *parent; // tevine klase
        TClass *next;   // klasiu sarashui (to pachio lygio)
        long    level;  // 0=root, 1=kitos, 2=tu vaikai, 3...
        // vtbl
        bool    v_ok;
        long    v_sz;
        long   *v_ptr;
        // visi stringai
        static TStringTable strings;
        // kitkas
        static long __id;
};

/***************************************************************************/
/* TString, TStringTable */

struct TString {
    TString( const CString& str )
        : s(str), next(NULL) { };
    CString     s;
    TString*    next;
};

class TStringTable {
public:
        TStringTable();
        ~TStringTable();
        // metodai
        long   find(const CString& s) const; // randa, grazina offseta arba -1
        long   add (const CString& s);       // prideda, grazina offseta
        // get
        long        Size() const { return size; };
        TString*    First() const { return first; };
private:
        TString *first;
        TString *last;
        long     size;
};

/***************************************************************************/
/* TType, TTypeTable */

enum ETypes { EUSER, ELONG, EFLOAT, EVOID, EREF, ESTR, EERROR, EMAX };

class TSymTable;

class TType : public TTableItem {
public:
        TType(const CString& n, ETypes     i); // builtin types
        TType(const CString& n, TType     *t); // typedef
        TType(const CString& n, TSymTable *m); // struct
        TType(const CString& n, TClass    *c); // class
        ~TType();
        // metodai
        void     print(long i, std::ostream& f);         // ishveda c++ def
        TSymbol *findSym(const CString& n) const;   // randa simb memberiuose
        // read
        TSymTable *Memb()    const { return members; };
        TClass    *Cla()     const { return cla; };
        TClass    *Belongs() const { return belongs; };
        long       Size()    const { return size; };
        ETypes     Id()      const { return id; };
        TType     *Ref()     const { return ref; };
        // write
        void Belongs(TClass *n) { belongs = n; };
private:
        TSymTable *members;     // strukturoms
        TClass    *cla;         // klasems
        TClass    *belongs;     // kuriai klasei priklauso
        long       size;        // dydis (4-baitais)
        ETypes     id;          // builtin tipams
        TType     *ref;         // ish tikro shitas (typedef)
};

class TTypeTable : public TTable {
public:
        TTypeTable( bool root = false );
        // methods
        void    add(TType *t) { TTable::addItem( t ); }
        TType  *find(const CString& n) const; // randa
        void    print(long i, std::ostream& f);    // ishveda c++ def
        static TType *ref(ETypes t) { return typeRefs[t]; };
private:
        static TType *typeRefs[EMAX];
};

#define TypeVoid  (TTypeTable::ref(EVOID))
#define TypeError (TTypeTable::ref(EERROR))
#define TypeLong  (TTypeTable::ref(ELONG))
#define TypeFloat (TTypeTable::ref(EFLOAT))
#define TypeRef   (TTypeTable::ref(EREF))
#define TypeStr   (TTypeTable::ref(ESTR))

/***************************************************************************/
/* TSymbol, TSymTable */

class TSymTable;

// simbolio klase
class TSymbol : public TTableItem {
public:
        // t-tipas, n-vardas, a-jei masyvas, s-jei static
        TSymbol(TType *t, const CString& n, long a = -1, bool s = false);
        ~TSymbol();
        // metodai
        void print(long i, std::ostream& f); // ishveda c++ def
        void print(TClass *c, const CString& vm, long i, std::ostream& f); // isveda static c++
        // read
        TType     *Type()   const { return type; };
        bool       Stat()   const { return stat; };
        long       Array()  const { return array; };
        long       Offset() const { return offset; };
        // write
        void Offset(long o)   { offset = o; };
private:
        TType   *type;   // tipas
        bool     stat;   // static
        long     offset; // offsetas nuo pradzhios
        long     array;  // jei >=0 - chia masyvas
};

class TSymTable : public TTable {
public:
        TSymTable(bool vptr = false);
        // metodai
        void add(TSymbol *s);
        void print(long i, std::ostream& f); // ishveda c++ def
        void print(TClass *c, const CString& vm, long i, std::ostream& f); // isveda static c++
        TSymbol *find(const CString& n) const
            { return (TSymbol*)TTable::findItem( n ); };
        // read
        long     Size()       const { return size; };
        long     Size(long n) const; // pirmu n simboliu
        long     Offset()     const { return offset; };
        // write
        void     Offset(long o);
private:
        long size;
        long offset;
};

/***************************************************************************/
/* TFunction, TFuncTable */

/* funkcijos gali buti:
   built-in (sin, cos, abs, ...) - static arba ne
   native (C++) arba BiTikZyz f-jos
*/

// builtin f-jos
enum EBuiltin {
    EB_NONE,
    // int fn(int)
    EB_RAND, EB_ABS, EB_SGN,
    // int fn(int, int)
    EB_MIN, EB_MAX,
    // int fn(int, int, int)
    EB_CLAMP,
    // float fn(float)
    EB_FABS, EB_FSGN, EB_SIN, EB_COS, EB_TAN, EB_ATAN, EB_EXP, EB_LOGE,
    EB_SQRT, EB_SQR,
    // float fn()
    EB_FRAND,
    // float fn(float, float)
    EB_FMIN, EB_FMAX,
    // float fn(float, float, float)
    EB_FCLAMP, EB_LINERP, EB_COSERP,
    // void fn(float)
    EB_PRINT,
};

struct TBuiltin {
    EBuiltin id;
    ETypes   ret;
    char    *name;
    ETypes  *args;
    bool     noref;
    char    *text;
};

struct TCallAddr {
    long addr;
    TSymbol *sym;
};

class TFunction : public TTableItem {
public:
        TFunction(TType *t, const CString& n, TSymTable *l);
        TFunction(TBuiltin *t); // builtin inline f-ja
        ~TFunction();
        // metodai
        void print(long i, std::ostream& f); // ishveda c++ def
        void print(TClass *c, long i, std::ostream& f); // ishveda c++ def
        void print(TClass *c, const CString& vm, long i, std::ostream& f); // isveda native c++
        void optimize();        // optimizuoja sintakses medi, tipu tikrin.
        void code();            // daro tarpini koda
        void opticode();        // optimizuoja tarpini koda
        void assemble(long b);  // daro vm asma (su base b)
        // read
        TType      *Type()    const { return type; };
        TSymTable  *Tab()     const { return tab; };
        long        Args()    const { return args; };
        EBuiltin    Builtin() const { return builtin; };
        bool        Noref()   const { return noref; };
        long        Native()  const { return native; };
        const CString& Text() const { return text; };
        long        Offset()  const { return offset; };
        TNode      *Tree()    const { return tree; };
        TIntInsn   *Op()      const { return op; };
        uchar      *As()      const { return as; };
        long        Size()    const { return size; };
        long        Base()    const { return base; };
        // write
        void Args(long a)       { args = a; };
        void Native(long n)     { native = n; };
        void Offset(long o)     { offset = o; };
        void Tree(TNode *n)     { tree = n; };
private:
        void fill();
        // f-jos info
        TType      *type;       // graz. tipas
        TSymTable  *tab;        // lokalus kint.
        long        args;       // arg. kiekis
        EBuiltin    builtin;    // gal jis inline f-ja ?
        bool        noref;      // gal jam nereikia ptr i objekta ?
        long        native;     // jei != -1: cia native slotas
        CString     text;       // builtin f-ju body tekstas
        long        offset;     // offsetas vtbl
        // kompiliuojant
        TNode      *tree;       // sintaks. medis
        TIntInsn   *op;         // tarpinis kodas
        // vm asmas
        uchar      *as;         // asmas
        long        size;       // asmo dydis baitais
        long        base;       // offsetas visam asme
};

class TFuncTable : public TTable {
public:
        TFuncTable(bool root = false);
        // metodai
        void add(TFunction *f); // prideda
        void replace(TFunction *f, TFunction *of); // overload vaikinese
        void print(long i, std::ostream& f); // ishveda c++ def
        void print(TClass *c, long i, std::ostream& f); // ishveda c++ def
        void print(TClass *c, const CString& vm, long i, std::ostream& f); // isveda native c++
        void optimize() const;
        void debug(long n) const;
        void code() const;
        TFunction *find(const CString& n) const
            { return (TFunction*)TTable::findItem( n ); }
        // read
        long       Size()  const { return size; };
        long       Base()  const { return base; };
        // write
        void Base(long b);
private:
        uchar     *as;          // gautas asmas
        long       asSize;      // dydis asmo
        long       size;        // kiek f-ju
        long       base;        // kiek nuo tevu prisikaupe
        static TBuiltin builtins[];
};

/***************************************************************************/
/* inlainai */

inline TType::~TType()
    { if (members) delete members; }
inline TSymbol::TSymbol(TType *t, const CString& n, long a, bool s)
    : TTableItem( n )
    { type = t; offset = 0; array = a; stat = s; }
inline TSymbol::~TSymbol()
    { }
inline void TFunction::fill()
{
    args=0; builtin=EB_NONE; noref=false; native=-1; tree=NULL;
    op=NULL; as=NULL; size=0; base=offset=0;
}
inline TFunction::TFunction(TType *t, const CString& n, TSymTable *s)
    : TTableItem( n )
    { fill(); type = t; tab = s; }

#endif

