// virtuali mashina

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef MMGR
#include "../utils/mmgr.h"
#endif

#include "vmachine.h"

#define VM_PI  3.14159265358979323846f // jei kartais math.h nepadaro :)

/** static hash table of strings ********************************************/

void TClassInfo::read(FILE* f)
{
    char buf[256];
    fread(&size, sizeof(long), 1, f);
    char c, l = 0;
    do { c = buf[l++] = fgetc(f); } while (c); // vardas iki 0 baito
    name = new char[l]; memcpy(name, buf, l);
    fread(&parent, sizeof(long), 1, f);
    fread(&nvtbl, sizeof(long), 1, f);
    vtbl = new long[nvtbl];
    fread(vtbl, sizeof(long), nvtbl, f);
}

// grazina klases id pagal varda
long TVMachine::find(char *n) const
{
    long i;
    for (i = 0; i < ncl; i++)
        if (!strcmp(n, cli[i].name))
             return i;
    return -1;
}
// grazina, ar klase A yra klase B
long TVMachine::is(long a, long b) const
{
    while (a >= 0) {
        if (a == b)
            return 1;
        a = cli[a].parent;
    }
    return 0;
}

// konstruktorius
TVMachine::TVMachine(char *fn, unsigned char *dt, TVMNative *na, long sz)
{
    long i;
    ok = false;
    // atidarom faila
    FILE* f = fopen(fn, "rb");
    if (!f) {
            printf("negaliu atidaryti failo %s\n", fn);
            return;
    }
    // skaitom koda
    fread(&ncode, sizeof(ncode), 1, f);
    code = new unsigned char[ncode + 4]; // 4 TIK KOLKAS
    fread(code, ncode, 1, f);
    // skaitom klases
    fread(&ncl, sizeof(ncl), 1, f);
    cli = new TClassInfo[ncl];
    for (i = 0; i < ncl; i++)
            cli[i].read(f);
    // skaitom stringus
    fread(&nstr, sizeof(nstr), 1, f);
    strings = new char[nstr];
    fread(strings, nstr, 1, f);
    // darom kitka
    stack = new unsigned char[sz];
    depth = 0;
    ssb = stack;
    data = dt;
    native = na;
    ok = true;
    fclose(f);
}

TVMachine::~TVMachine()
{
    long i;
    if (!ok)
            return;
    for (i = 0; i < ncl; i++) {
            if (cli[i].name ) delete[] cli[i].name ;
            if (cli[i].vtbl ) delete[] cli[i].vtbl ;
    }
    delete[] cli;
    delete[] code;
    delete[] stack;
}

// KRUVA DEFINU

#define VMSGN(A) ((A)>=0)?(1):(-1)
#define VMSGNf(A) ((A)>=0)?(1.0f):(-1.0f)
#define VMMIN(A, B) ((A)<(B))?(A):(B)
#define VMMAX(A, B) ((A)<(B))?(A):(B)

#define COUNT ((*(long*)(ip-4))>>16)

// kviechia i klases j f-ja
void TVMachine::call(long i, long j)
{
    depth++;
    long initDepth = depth;
    //unsigned char *ip;
    ip = code + cli[i].vtbl[j]; // jmp
    _pushp(ip); // ret. adr
    sbase[depth] = ssb; // frame-ptr
    do {
        // nuskaitom insn opkoda
        EIntOpcode op = (EIntOpcode)((*(long*)ip)&0xFFFF);
        ip += 4;
        // dekodinam op
        switch (op) {
        case IOP_NOP:   break;
        // store/load
        case IOP_SAVE:  _save(COUNT); break;
        case IOP_PUSH:  _push(COUNT); break;
        // ptr arith
        case IOP_REFFR:   _pushp(sbase[depth]); break;
        case IOP_REFTHIS: _pushp(_getthis()); break;
        case IOP_REFSTAT: _pushp(data); break;
        case IOP_REFADD:  sspi[-2] += ssi[-1]; ssi--; break;
        // steko dydzhio keitimas
        case IOP_DEL:   ssi -= COUNT; break;
        case IOP_GROW:  ssi += COUNT; break;
        // call/callsuper/native/ret
        case IOP_CALL: {
                // push ret
                _pushp(ip);
                sbase[++depth] = ssb; // naujas stekas
                ip = code + cli[*sspi[-2]].vtbl[COUNT];
                } break;
        case IOP_CALLSUPER: {
                // push ret
                _pushp(ip + 4);
                sbase[++depth] = ssb; // naujas stekas
                ip = code + cli[*(long*)ip].vtbl[COUNT];
                } break;
        case IOP_NATIVE: { // kvieciam wraperi, kuris kviecia C++
                unsigned char *tmp = ip;
                (native[COUNT])((long*)sbase[depth]);
                ip = tmp;
                } break;
        case IOP_RET:
                ssi--;  ip = sspb[0];
                depth--;        // senas stekas
                if (depth < initDepth) goto _viskas; // grazu...
                break;
        // new/delete
        case IOP_NEW: {
                long c = COUNT;
                _pushp(create(c));
                } break;
        case IOP_IS:
                ssi[-1] = is(*((long*)(*(sspi[-1]))), COUNT);
                break;
        case IOP_DELETE: {
                long *p = sspi[-1]; sspi--;
                delete[] p;
                } break;
        // push const
        case IOP_PUSHs:   _pushi(((long)strings) + COUNT); break;
        case IOP_PUSHif:  _pushi(*(long *)ip); ip += 4; break;
        case IOP_PUSHarr: _pushArr(COUNT); break;
        // konversijos
        case IOP_i2f: ssf[-1] = (float)(ssi[-1]); break;
        case IOP_f2i: ssi[-1] = (long )(ssf[-1]); break;
        // jmp
        case IOP_JMP:   ip = code + (*(long*)ip); break;
        case IOP_JZ:
                ssi--;
                if (!ssi[0])
                        ip = code + (*(long*)ip);
                else
                        ip += 4;
                break;
        // 2xlong arith
        case IOP_OR:    ssi[-2] |=  ssi[-1]; ssi--; break;
        case IOP_XOR:   ssi[-2] ^=  ssi[-1]; ssi--; break;
        case IOP_AND:   ssi[-2] &=  ssi[-1]; ssi--; break;
        case IOP_SHL:   ssi[-2] <<= ssi[-1]; ssi--; break;
        case IOP_SHR:   ssi[-2] >>= ssi[-1]; ssi--; break;
        case IOP_MOD:   ssi[-2] %=  ssi[-1]; ssi--; break;
        case IOP_ADDi:  ssi[-2] +=  ssi[-1]; ssi--; break;
        case IOP_SUBi:  ssi[-2] -=  ssi[-1]; ssi--; break;
        case IOP_MULi:  ssi[-2] *=  ssi[-1]; ssi--; break;
        case IOP_DIVi:  ssi[-2] /=  ssi[-1]; ssi--; break;
        // 2xlong bool arith
        case IOP_BORi:  ssi[-2] = ssi[-2] || ssi[-1]; ssi--; break;
        case IOP_BANDi: ssi[-2] = ssi[-2] && ssi[-1]; ssi--; break;
        case IOP_BEi:   ssi[-2] = ssi[-2] == ssi[-1]; ssi--; break;
        case IOP_BNEi:  ssi[-2] = ssi[-2] != ssi[-1]; ssi--; break;
        case IOP_BLi:   ssi[-2] = ssi[-2] <  ssi[-1]; ssi--; break;
        case IOP_BGi:   ssi[-2] = ssi[-2] >  ssi[-1]; ssi--; break;
        case IOP_BLEi:  ssi[-2] = ssi[-2] <= ssi[-1]; ssi--; break;
        case IOP_BGEi:  ssi[-2] = ssi[-2] >= ssi[-1]; ssi--; break;
        // 2xfloat arith
        case IOP_ADDf:  ssf[-2] += ssf[-1]; ssf--; break;
        case IOP_SUBf:  ssf[-2] -= ssf[-1]; ssf--; break;
        case IOP_MULf:  ssf[-2] *= ssf[-1]; ssf--; break;
        case IOP_DIVf:  ssf[-2] /= ssf[-1]; ssf--; break;
        // 2xfloat bool arith
        case IOP_BORf:  ssi[-2] = ssf[-2] || ssf[-1]; ssi--; break;
        case IOP_BANDf: ssi[-2] = ssf[-2] && ssf[-1]; ssi--; break;
        case IOP_BEf:   ssi[-2] = ssf[-2] == ssf[-1]; ssi--; break;
        case IOP_BNEf:  ssi[-2] = ssf[-2] != ssf[-1]; ssi--; break;
        case IOP_BLf:   ssi[-2] = ssf[-2] <  ssf[-1]; ssi--; break;
        case IOP_BGf:   ssi[-2] = ssf[-2] >  ssf[-1]; ssi--; break;
        case IOP_BLEf:  ssi[-2] = ssf[-2] <= ssf[-1]; ssi--; break;
        case IOP_BGEf:  ssi[-2] = ssf[-2] >= ssf[-1]; ssi--; break;
        // 1xlong arith
        case IOP_NEGi:  ssi[-1] = -ssi[-1]; break;
        case IOP_BNOTi: ssi[-1] = !ssi[-1]; break;
        case IOP_INV:   ssi[-1] = ~ssi[-1]; break;
        // long arith assign
        case IOP_SAVEMULi: *(sspi[-1]) *=  ssi[-2]; ssi--; break;
        case IOP_SAVEDIVi: *(sspi[-1]) /=  ssi[-2]; ssi--; break;
        case IOP_SAVEMOD:  *(sspi[-1]) %=  ssi[-2]; ssi--; break;
        case IOP_SAVEADDi: *(sspi[-1]) +=  ssi[-2]; ssi--; break;
        case IOP_SAVESUBi: *(sspi[-1]) -=  ssi[-2]; ssi--; break;
        case IOP_SAVESHL:  *(sspi[-1]) <<= ssi[-2]; ssi--; break;
        case IOP_SAVESHR:  *(sspi[-1]) >>= ssi[-2]; ssi--; break;
        case IOP_SAVEOR:   *(sspi[-1]) |=  ssi[-2]; ssi--; break;
        case IOP_SAVEXOR:  *(sspi[-1]) ^=  ssi[-2]; ssi--; break;
        case IOP_SAVEAND:  *(sspi[-1]) &=  ssi[-2]; ssi--; break;
        // 1xfloat arith
        case IOP_NEGf:  ssf[-1] = -ssf[-1]; break;
        case IOP_BNOTf: ssi[-1] = !ssf[-1]; break;
        // float arith assign
        case IOP_SAVEMULf: *(sspf[-1]) *= ssf[-2]; ssf--; break;
        case IOP_SAVEDIVf: *(sspf[-1]) /= ssf[-2]; ssf--; break;
        case IOP_SAVEADDf: *(sspf[-1]) += ssf[-2]; ssf--; break;
        case IOP_SAVESUBf: *(sspf[-1]) -= ssf[-2]; ssf--; break;
        // inc/dec
        //   taip VC buginasi, todel hacka darau :)
        //case IOP_PREINCi: ssi[-1] = ++(*(sspi[-1]))  ; break;
        //case IOP_PREDECi: ssi[-1] = --(*(sspi[-1]))  ; break;
        //case IOP_POSINCi: ssi[-1] =   (*(sspi[-1]))++; break;
        //case IOP_POSDECi: ssi[-1] =   (*(sspi[-1]))--; break;
        //case IOP_PREINCf: ssf[-1] = ++(*(sspf[-1]))  ; break;
        //case IOP_PREDECf: ssf[-1] = --(*(sspf[-1]))  ; break;
        //case IOP_POSINCf: ssf[-1] =   (*(sspf[-1]))++; break;
        //case IOP_POSDECf: ssf[-1] =   (*(sspf[-1]))--; break;
        case IOP_PREINCi: { long* p = sspi[-1]; ssi[-1] = ++(*p); } break;
        case IOP_PREDECi: { long* p = sspi[-1]; ssi[-1] = --(*p); } break;
        case IOP_POSINCi: { long* p = sspi[-1]; ssi[-1] = (*p)++; } break;
        case IOP_POSDECi: { long* p = sspi[-1]; ssi[-1] = (*p)--; } break;
        case IOP_PREINCf: { float* p = sspf[-1]; ssf[-1] = ++(*p); } break;
        case IOP_PREDECf: { float* p = sspf[-1]; ssf[-1] = --(*p); } break;
        case IOP_POSINCf: { float* p = sspf[-1]; ssf[-1] = (*p)++; } break;
        case IOP_POSDECf: { float* p = sspf[-1]; ssf[-1] = (*p)--; } break;
        // advanced
        case IOP_MULADDf: ssf[-3] += ssf[-2] * ssf[-1]; ssf--; break;
        case IOP_MULDIVf: ssf[-3] /= ssf[-2] * ssf[-1]; ssf--; break;
        case IOP_POP: {
                long c = COUNT;
                _save(c); ssi -= c;
                } break;
        case IOP_OFSTHIS:
                _pushp(_getthis() + *(long*)ip);
                ip += 4; break;
        case IOP_OFSFR:
                _pushp((long*)(sbase[depth]) + *(long*)ip);
                ip += 4; break;
        case IOP_OFSPTR:
                sspi[-1] = *(long**)(sspi[-1]) + *(long*)ip;
                ip += 4; break;
        case IOP_SAVE1: ssi--; *sspi[0] = ssi[-1]; break;
        case IOP_PUSH1: ssi[-1] = *sspi[-1]; break;
        case IOP_POP1: ssi-=2; *sspi[1] = ssi[0]; break;
        case IOP_DEL1: ssi--; break;
        // inline "f-jos"
        case IOP_RAND: ssi[-1] = rand() % (ssi[-1]); break;
        case IOP_ABS:  ssi[-1] = abs(ssi[-1]); break;
        case IOP_SGN:  ssi[-1] = VMSGN(ssi[-1]); break;
        case IOP_MIN: ssi[-2] = VMMIN(ssi[-2], ssi[-1]); ssi--; break;
        case IOP_MAX: ssi[-2] = VMMAX(ssi[-2], ssi[-1]); ssi--; break;
        case IOP_CLAMP: {
                long t = ssi[-3];
                if (t < ssi[-2])      t = ssi[-2];
                else if (t > ssi[-1]) t = ssi[-1];
                ssi[-2] = t;         ssi -= 2;
                } break;
        case IOP_FABS: ssf[-1] = fabsf (ssf[-1]); break;
        case IOP_FSGN: ssf[-1] = VMSGNf(ssf[-1]); break;
        case IOP_SIN:  ssf[-1] = sinf  (ssf[-1]); break;
        case IOP_COS:  ssf[-1] = cosf  (ssf[-1]); break;
        case IOP_TAN:  ssf[-1] = tanf  (ssf[-1]); break;
        case IOP_ATAN: ssf[-1] = atanf (ssf[-1]); break;
        case IOP_EXP:  ssf[-1] = expf  (ssf[-1]); break;
        case IOP_LOGE: ssf[-1] = logf  (ssf[-1]); break;
        case IOP_SQRT: ssf[-1] = sqrtf (ssf[-1]); break;
        case IOP_SQR:  ssf[-1] = ssf[-1] * ssf[-1]; break;
        case IOP_FRAND: ssf[0] = (float)(rand())/RAND_MAX; ssf++; break;
        case IOP_FMIN:  ssf[-2] = VMMIN(ssf[-2], ssf[-1]); ssf--; break;
        case IOP_FMAX:  ssf[-2] = VMMAX(ssf[-2], ssf[-1]); ssf--; break;
        case IOP_FCLAMP: {
                float t = ssf[-3];
                if (t < ssf[-2])      t = ssf[-2];
                else if (t > ssf[-1]) t = ssf[-1];
                ssf[-3] = t;         ssf -= 2;
                } break;
        case IOP_LINERP: ssf[-3] += ssf[-1] * (ssf[-2] - ssf[-3]);
                ssf -= 2; break;
        case IOP_COSERP: {
                float ftmp = (1 - cosf(ssf[-1] * VM_PI)) * 0.5f;
                ssf[-3] += ftmp * (ssf[-2] - ssf[-3]);  ssf -= 2;
                } break;
        case IOP_PRINT:
                printf("  %f\n", (double)(ssf[-1])); ssf--;
                break;
        case JUMPTARGET: // foo, kad warningo nebutu
                break;
        }
    } while (1);
   _viskas:
    return;
}

